# Copyright 2023 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from urllib.parse import urlparse

import clickhouse_driver
from flask import abort

from . import consts
from . import access
from .models import Branch, Flow, Run, Job, Agent


def get_clickhouse_url():
    ch_url = os.environ.get('KRAKEN_CLICKHOUSE_URL', consts.DEFAULT_CLICKHOUSE_URL)
    return ch_url


def get_clickhouse():
    ch_url = get_clickhouse_url()
    o = urlparse(ch_url)
    ch = clickhouse_driver.Client(host=o.hostname)
    return ch


def prepare_logs_query(branch_id=None, flow_kind=None, flow_id=None, run_id=None, job_id=None, step_idx=None,
                       agent_id=None, services=None, level=None, token_info=None):
    if branch_id is None and flow_id is None and run_id is None and job_id is None and agent_id is None:
        access.check(token_info, '', 'admin',
                     'only superadmin can get services logs')

    if branch_id is None and flow_kind is not None:
        abort(400, "if flow_kind is provided then branch_id must be provided as well")

    if job_id is None and step_idx is not None:
        abort(400, "if step_idx is provided then job_id must be provided as well")

    cols_to_skip = set()

    if step_idx is not None:
        cols_to_skip.add('step_idx')

    if job_id is not None:
        job = Job.query.filter_by(id=job_id).one_or_none()
        if job is None:
            abort(404, "Job not found")
        access.check(token_info, job.run.stage.branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get job logs')
        cols_to_skip.add('job')
        cols_to_skip.add('run')
        cols_to_skip.add('flow')
        cols_to_skip.add('flow_kind')
        cols_to_skip.add('branch')

    elif run_id is not None:
        run = Run.query.filter_by(id=run_id).one_or_none()
        if run is None:
            abort(404, "Run not found")
        access.check(token_info, run.stage.branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get run logs')
        cols_to_skip.add('run')
        cols_to_skip.add('flow')
        cols_to_skip.add('flow_kind')
        cols_to_skip.add('branch')

    elif flow_id is not None:
        flow = Flow.query.filter_by(id=flow_id).one_or_none()
        if flow is None:
            abort(404, "Flow not found")
        access.check(token_info, flow.branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get flow logs')
        cols_to_skip.add('flow')
        cols_to_skip.add('flow_kind')
        cols_to_skip.add('branch')

    elif branch_id is not None:
        branch = Branch.query.filter_by(id=branch_id).one_or_none()
        if branch is None:
            abort(404, "Branch not found")
        access.check(token_info, branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get branch logs')
        cols_to_skip.add('branch')

    if agent_id is not None:
        agent = Agent.query.filter_by(id=agent_id).one_or_none()
        if agent is None:
            abort(404, "Agent not found")
        access.check(token_info, '', 'admin',
                     'only superadmin can get agent logs')
        cols_to_skip.add('agent')

    params = {}

    where_clauses = []
    if branch_id is not None:
        where_clauses.append('branch = %(branch_id)d')
        params['branch_id'] = branch_id
        if flow_kind is not None:
            where_clauses.append('flow_kind = %(flow_kind)d')
            params['flow_kind'] = flow_kind
    if flow_id is not None:
        where_clauses.append('flow = %(flow_id)d')
        params['flow_id'] = flow_id
    if run_id is not None:
        where_clauses.append('run = %(run_id)d')
        params['run_id'] = run_id
    if job_id is not None:
        where_clauses.append('job = %(job_id)d')
        params['job_id'] = job_id
        if step_idx is not None:
            where_clauses.append('step_idx = %(step_idx)d')
            params['step_idx'] = step_idx
    if agent_id is not None:
        where_clauses.append('agent = %(agent_id)d')
        params['agent_id'] = agent_id

    where_services = []
    if services:
        cols_to_skip.add('agent')
        if len(services) == 1:
            cols_to_skip.add('service')

        for idx, s in enumerate(services):
            param = 'service%d' % idx
            if '/' in s:
                s, t = s.split('/')
                tparam = 'tool%d' % idx
                where_services.append("(service = %%(%s)s and tool = %%(%s)s)" % (param, tparam))
                params[param] = s
                params[tparam] = t
            else:
                where_services.append("service = %%(%s)s" % param)
                params[param] = s
        where = " or ".join(where_services)
        where = "(" + where + ") "
        where_clauses.append(where)

    if level:
        level = level.upper()
        if level == 'ERROR':
            lq = "level = 'ERROR'"
        elif level == 'WARNING':
            lq = "level in ('WARNING', 'ERROR')"
        else:
            lq = "level in ('INFO', 'WARNING', 'ERROR')"
        where_clauses.append(lq)

    if where_clauses:
        where_clause = 'where ' + ' and '.join(where_clauses)
    else:
        where_clause = ''

    columns = ['time', 'message', 'service', 'host', 'path', 'lineno', 'level',
               'branch', 'flow_kind','flow', 'run', 'job', 'tool', 'step', 'agent']
    for col in cols_to_skip:
        columns.remove(col)
    columns = ','.join(columns)

    return columns, where_clause, params



"""added env_vars to branch

Revision ID: 68293032700d
Revises: 9e57e0f30a5f
Create Date: 2023-07-24 07:41:57.915270

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '68293032700d'
down_revision = '9e57e0f30a5f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('branches', sa.Column('env_vars', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('branches', 'env_vars')
    # ### end Alembic commands ###



# Copyright 2020-2023 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from queue import Queue, Empty
from threading import Thread, Event

from flask import request, abort, Response
from . import access
from . import users
from .models import Job
from . import chops
from . import utils


log = logging.getLogger(__name__)

class LogDownloader:
    def __init__(self, columns, query, params=None):
        self.columns = columns
        self.query = query
        self.params = params
        self.ch = chops.get_clickhouse()
        self.logs_queue = Queue()
        self.finished = Event()
        self.worker = None

    def get_logs(self):
        self.logs_queue.put(self.columns + '\n')
        for line_items in self.ch.execute_iter(self.query, self.params, settings={'max_block_size': 100000}):
            line = ' '.join([str(i) for i in line_items])
            self.logs_queue.put(line + '\n')

        self.logs_queue.join()  # wait for all blocks in the queue to be marked as processed
        self.finished.set()  # mark streaming as finished

    def send_logs(self):
        while not self.finished.is_set():
            try:
                yield self.logs_queue.get(timeout=0.01)
                self.logs_queue.task_done()
            except Empty:
                self.finished.wait(0.01)
        self.worker.join()

    def download(self, ):
        self.worker = Thread(target=self.get_logs)
        self.worker.start()
        return self.send_logs()


def serve_job_log(job_id):
    job_id = int(job_id)
    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")

    token_info = users.get_token_info_from_request()
    if not token_info:
        abort(401, "Missing Authorization header or kk_session_token cookie")
    access.check(token_info, job.run.flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can fetch job logs')

    query = 'SELECT message FROM logs WHERE job = %d ' % job_id
    query += 'ORDER BY time ASC, seq ASC'

    try:
        jld = LogDownloader('message', query)
    except OSError:
        abort(500, 'Cannot connect to storage service')

    log.info('serve_job_log %s', job)
    response = Response(jld.download(), mimetype='plain/text')
    path = 'job_log_%d.txt' % job.id
    response.headers['Content-Disposition'] = 'attachment; filename=' + path
    return response


def serve_step_log(job_id, step_idx):
    job_id = int(job_id)
    step_idx = int(step_idx)
    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")

    token_info = users.get_token_info_from_request()
    if not token_info:
        abort(401, "Missing Authorization header or kk_session_token cookie")
    access.check(token_info, job.run.flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can fetch job logs')

    query = 'SELECT message FROM logs WHERE job = %d ' % job_id
    query += ' AND step = %d ' % step_idx
    query += 'ORDER BY time ASC, seq ASC'

    try:
        jld = LogDownloader('message', query)
    except OSError:
        abort(500, 'Cannot connect to storage service')

    log.info('serve_job_log %s', job)
    response = Response(jld.download(), mimetype='plain/text')
    path = 'step_log_%d_%d.txt' % (job_id, step_idx)
    response.headers['Content-Disposition'] = 'attachment; filename=' + path
    return response


def serve_any_log():
    token_info = users.get_token_info_from_request()
    branch_id = request.args.get('branch_id', default=None, type=int)
    flow_kind = request.args.get('flow_kind', default=None, type=int)
    flow_id = request.args.get('flow_id', default=None, type=int)
    run_id = request.args.get('run_id', default=None, type=int)
    job_id = request.args.get('job_id', default=None, type=int)
    step_idx = request.args.get('step_id', default=None, type=int)
    agent_id = request.args.get('agent_id', default=None, type=int)
    services = request.args.getlist('services')
    level = request.args.get('level', default=None, type=int)

    columns, where_clause, params = chops.prepare_logs_query(
        branch_id, flow_kind, flow_id, run_id, job_id, step_idx,
        agent_id, services, level, token_info)

    query = f'SELECT {columns} FROM logs {where_clause} '
    query += 'ORDER BY time ASC, seq ASC'

    log.info('log query: %s', query)

    try:
        jld = LogDownloader(columns, query, params)
    except OSError:
        abort(500, 'Cannot connect to storage service')

    # log.info('kraken_log %s', job)
    response = Response(jld.download(), mimetype='plain/text')
    now = utils.utcnow().strftime("%Y_%m_%d_%H_%M_%S")
    path = 'kraken_log_%s.txt' % now
    response.headers['Content-Disposition'] = 'attachment; filename=' + path
    return response



"""added user_data to several entities

Revision ID: bdcdbb4b5b3f
Revises: cea0e48abb5a
Create Date: 2023-03-08 19:55:37.167491

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'bdcdbb4b5b3f'
down_revision = 'cea0e48abb5a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('branches', sa.Column('user_data', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('branches', sa.Column('user_data_ci', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('branches', sa.Column('user_data_dev', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('flows', sa.Column('user_data', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('projects', sa.Column('user_data', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('projects', 'user_data')
    op.drop_column('flows', 'user_data')
    op.drop_column('branches', 'user_data_dev')
    op.drop_column('branches', 'user_data_ci')
    op.drop_column('branches', 'user_data')
    # ### end Alembic commands ###
